@@ -1,5 +1,5 @@
-#This file is part of Tryton.  The COPYRIGHT file at the top level of
-#this repository contains the full copyright notices and license terms.
+# This file is part of Tryton.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
 from shlex import shlex
 from types import GeneratorType
 import gettext
@@ -7,12 +7,13 @@ import locale
 import decimal
 from decimal import Decimal
 import datetime
-import time
 import io
 from collections import OrderedDict
 
-from tryton.translate import date_format
 from tryton.common import untimezoned_date, timezoned_date, datetime_strftime
+from tryton.common.datetime_ import date_parse
+from tryton.common.timedelta import parse as timedelta_parse
+from tryton.common.timedelta import format as timedelta_format
 from tryton.pyson import PYSONDecoder
 
 __all__ = ['DomainParser']
@@ -41,8 +42,8 @@ class udlex(shlex):
         class DummyWordchars(object):
             "Simulate str that contains all chars except somes"
             def __contains__(self, item):
-                return item not in (u':', u'>', u'<', u'=', u'!', u'"', u';',
-                    u'(', u')')
+                return item not in (':', '>', '<', '=', '!', '"', ';',
+                    '(', ')')
 
         self.wordchars = DummyWordchars()
 
@@ -80,7 +81,10 @@ def simplify(value):
 
 def group_operator(tokens):
     "Group token of operators"
-    cur = next(tokens)
+    try:
+        cur = next(tokens)
+    except StopIteration:
+        return
     nex = None
     for nex in tokens:
         if nex == '=' and cur and cur + nex in OPERATORS:
@@ -115,7 +119,7 @@ def likify(value):
 
 def quote(value):
     "Quote string if needed"
-    if not isinstance(value, basestring):
+    if not isinstance(value, str):
         return value
     if '\\' in value:
         value = value.replace('\\', '\\\\')
@@ -157,7 +161,7 @@ def append_ending_clause(domain, clause, deep):
     "Append clause after the ending clause"
     if not domain:
         yield clause
-        raise StopIteration
+        return
     for dom in domain[:-1]:
         yield dom
     if isinstance(domain[-1], list):
@@ -195,7 +199,7 @@ def split_target_value(field, value):
     "Split the reference value into target and value"
     assert field['type'] == 'reference'
     target = None
-    if isinstance(value, basestring):
+    if isinstance(value, str):
         for key, text in field['selection']:
             if value.lower().startswith(text.lower() + ','):
                 target = key
@@ -223,67 +227,77 @@ def test_split_target_value():
         assert split_target_value(field, value) == result
 
 
-def convert_value(field, value):
+def convert_value(field, value, context=None):
     "Convert value for field"
+    if context is None:
+        context = {}
 
     def convert_boolean():
-        if isinstance(value, basestring):
-            return any(test.decode('utf-8').lower().startswith(value.lower())
-                for test in (_('y'), _('yes'), _('true'), _('t'), '1'))
+        if isinstance(value, str):
+            return any(test.lower().startswith(value.lower())
+                for test in (
+                    _('y'), _('Yes'), _('True'), _('t'), '1'))
         else:
             return bool(value)
 
     def convert_float():
+        factor = float(field.get('factor', 1))
         try:
-            return locale.atof(value)
+            return locale.atof(value) / factor
         except (ValueError, AttributeError):
             return
 
     def convert_integer():
+        factor = float(field.get('factor', 1))
         try:
-            return int(locale.atof(value))
+            return int(locale.atof(value) / factor)
         except (ValueError, AttributeError):
             return
 
     def convert_numeric():
+        factor = Decimal(field.get('factor', 1))
         try:
-            return locale.atof(value, Decimal)
+            return locale.atof(value, Decimal) / factor
         except (decimal.InvalidOperation, AttributeError):
             return
 
     def convert_selection():
-        if isinstance(value, basestring):
+        if isinstance(value, str):
             for key, text in field['selection']:
                 if value.lower() == text.lower():
                     return key
         return value
 
     def convert_datetime():
+        if not value:
+            return
+        format_ = context.get('date_format', '%x') + ' %X'
         try:
-            return untimezoned_date(datetime.datetime(*time.strptime(value,
-                        date_format() + ' ' + time_format(field))[:6]))
+            dt = date_parse(value, format_)
+            return untimezoned_date(dt)
         except ValueError:
-            try:
-                return datetime.datetime(*time.strptime(value,
-                        date_format())[:6])
-            except ValueError:
-                return
-        except TypeError:
             return
 
     def convert_date():
+        if not value:
+            return
+        format_ = context.get('date_format', '%x')
         try:
-            return datetime.date(*time.strptime(value, date_format())[:3])
+            return date_parse(value, format_).date()
         except (ValueError, TypeError):
             return
 
     def convert_time():
+        if not value:
+            return
         try:
-            return datetime.time(*time.strptime(value,
-                    time_format(field))[3:6])
+            return date_parse(value).time()
         except (ValueError, TypeError):
             return
 
+    def convert_timedelta():
+        return timedelta_parse(value, context.get(field.get('converter')))
+
     def convert_many2one():
         if value == '':
             return None
@@ -299,6 +313,7 @@ def convert_value(field, value):
         'datetime': convert_datetime,
         'date': convert_date,
         'time': convert_time,
+        'timedelta': convert_timedelta,
         'many2one': convert_many2one,
         }
     return converts.get(field['type'], lambda: value)()
@@ -336,6 +351,14 @@ def test_convert_float():
         assert convert_value(field, value) == result
 
 
+def test_convert_float_factor():
+    field = {
+        'type': 'float',
+        'factor': '100',
+        }
+    assert convert_value(field, '42') == 0.42
+
+
 def test_convert_integer():
     field = {
         'type': 'integer',
@@ -350,6 +373,14 @@ def test_convert_integer():
         assert convert_value(field, value) == result
 
 
+def test_convert_integer_factor():
+    field = {
+        'type': 'integer',
+        'factor': '2',
+        }
+    assert convert_value(field, '6') == 3
+
+
 def test_convert_numeric():
     field = {
         'type': 'numeric',
@@ -364,6 +395,14 @@ def test_convert_numeric():
         assert convert_value(field, value) == result
 
 
+def test_convert_numeric_factor():
+    field = {
+        'type': 'numeric',
+        'factor': '5',
+        }
+    assert convert_value(field, '1') == Decimal('0.2')
+
+
 def test_convert_selection():
     field = {
         'type': 'selection',
@@ -392,13 +431,17 @@ def test_convert_datetime():
         'format': '"%H:%M:%S"',
         }
     for value, result in (
-            ('12/04/2002', datetime.datetime(2002, 12, 4)),
+            ('12/04/2002', untimezoned_date(datetime.datetime(2002, 12, 4))),
             ('12/04/2002 12:30:00', untimezoned_date(
                     datetime.datetime(2002, 12, 4, 12, 30))),
+            ('02/03/04', untimezoned_date(datetime.datetime(2004, 2, 3))),
+            ('02/03/04 05:06:07', untimezoned_date(
+                    datetime.datetime(2004, 2, 3, 5, 6, 7))),
             ('test', None),
             (None, None),
             ):
-        assert convert_value(field, value) == result
+        assert convert_value(field, value) == result, (value,
+            convert_value(field, value), result)
 
 
 def test_convert_date():
@@ -426,15 +469,30 @@ def test_convert_time():
         assert convert_value(field, value) == result
 
 
-def format_value(field, value, target=None):
+def test_convert_timedelta():
+    field = {
+        'type': 'timedelta',
+        }
+    for value, result in [
+            ('1d 2:00', datetime.timedelta(days=1, hours=2)),
+            ('foo', datetime.timedelta()),
+            (None, None),
+            ]:
+        assert convert_value(field, value) == result
+
+
+def format_value(field, value, target=None, context=None):
     "Format value for field"
+    if context is None:
+        context = {}
 
     def format_boolean():
         return _('True') if value else _('False')
 
     def format_integer():
+        factor = float(field.get('factor', 1))
         if value or value is 0 or isinstance(value, float):
-            return str(int(value))
+            return str(int(value * factor))
         return ''
 
     def format_float():
@@ -442,11 +500,16 @@ def format_value(field, value, target=None):
                 and value is not 0
                 and not isinstance(value, (float, Decimal))):
             return ''
+        if isinstance(value, Decimal):
+            cast = Decimal
+        else:
+            cast = float
+        factor = cast(field.get('factor', 1))
         try:
-            digit = len(str(value).split('.')[1])
+            digit = len(str(value * factor).rstrip('0').split('.')[1])
         except IndexError:
             digit = 0
-        return locale.format('%.*f', (digit, value or 0), True)
+        return locale.format('%.*f', (digit, value * factor or 0), True)
 
     def format_selection():
         selections = dict(field['selection'])
@@ -461,25 +524,31 @@ def format_value(field, value, target=None):
     def format_datetime():
         if not value:
             return ''
-        format_ = date_format() + ' ' + time_format(field)
-        if (not isinstance(value, datetime.datetime)
-                or value.time() == datetime.time.min):
-            format_ = date_format()
-            time = value
+        format_ = context.get('date_format', '%x') + ' ' + time_format(field)
+        if not isinstance(value, datetime.datetime):
+            time = datetime.datetime.combine(value, datetime.time.min)
         else:
             time = timezoned_date(value)
+        if time.time() == datetime.time.min:
+            format_ = '%x'
         return datetime_strftime(time, format_)
 
     def format_date():
         if not value:
             return ''
-        return datetime_strftime(value, date_format())
+        format_ = context.get('date_format', '%x')
+        return datetime_strftime(value, format_)
 
     def format_time():
         if not value:
             return ''
         return datetime.time.strftime(value, time_format(field))
 
+    def format_timedelta():
+        if not value:
+            return ''
+        return timedelta_format(value, context.get(field.get('converter')))
+
     def format_many2one():
         if value is None:
             return ''
@@ -495,10 +564,11 @@ def format_value(field, value, target=None):
         'datetime': format_datetime,
         'date': format_date,
         'time': format_time,
+        'timedelta': format_timedelta,
         'many2one': format_many2one,
         }
     if isinstance(value, (list, tuple)):
-        return ';'.join(format_value(field, x) for x in value)
+        return ';'.join(format_value(field, x, context=context) for x in value)
     return quote(converts.get(field['type'],
             lambda: value if value is not None else '')())
 
@@ -530,6 +600,14 @@ def test_format_integer():
         assert format_value(field, value) == result
 
 
+def test_format_integer_factor():
+    field = {
+        'type': 'integer',
+        'factor': '2',
+        }
+    assert format_value(field, 3) == '6'
+
+
 def test_format_float():
     field = {
         'type': 'float',
@@ -540,13 +618,21 @@ def test_format_float():
             (1.50, '1.5'),
             (150.79, '150.79'),
             (0, '0'),
-            (0.0, '0.0'),
+            (0.0, '0'),
             (False, ''),
             (None, ''),
             ):
         assert format_value(field, value) == result
 
 
+def test_format_float_factor():
+    field = {
+        'type': 'float',
+        'factor': '100',
+        }
+    assert format_value(field, 0.42) == '42'
+
+
 def test_format_numeric():
     field = {
         'type': 'numeric',
@@ -554,16 +640,24 @@ def test_format_numeric():
     for value, result in (
             (Decimal(1), '1'),
             (Decimal('1.5'), '1.5'),
-            (Decimal('1.50'), '1.50'),
+            (Decimal('1.50'), '1.5'),
             (Decimal('150.79'), '150.79'),
             (Decimal(0), '0'),
-            (Decimal('0.0'), '0.0'),
+            (Decimal('0.0'), '0'),
             (False, ''),
             (None, ''),
             ):
         assert format_value(field, value) == result
 
 
+def test_format_numeric_factor():
+    field = {
+        'type': 'numeric',
+        'factor': '5',
+        }
+    assert format_value(field, Decimal('0.2')) == '1'
+
+
 def test_format_selection():
     field = {
         'type': 'selection',
@@ -593,7 +687,7 @@ def test_format_datetime():
         }
     for value, result in (
             (datetime.date(2002, 12, 4), '12/04/2002'),
-            (datetime.datetime(2002, 12, 4), '12/04/2002'),
+            (untimezoned_date(datetime.datetime(2002, 12, 4)), '12/04/2002'),
             (untimezoned_date(datetime.datetime(2002, 12, 4, 12, 30)),
                 '"12/04/2002 12:30:00"'),
             (False, ''),
@@ -627,6 +721,19 @@ def test_format_time():
         assert format_value(field, value) == result
 
 
+def test_format_timedelta():
+    field = {
+        'type': 'timedelta',
+        }
+    for value, result in [
+            (datetime.timedelta(days=1, hours=2), '"1d 02:00"'),
+            (datetime.timedelta(), ''),
+            (None, ''),
+            ('', ''),
+            ]:
+        assert format_value(field, value) == result
+
+
 def complete_value(field, value):
     "Complete value for field"
 
@@ -639,7 +746,7 @@ def complete_value(field, value):
     def complete_selection():
         test_value = value if value is not None else ''
         if isinstance(value, list):
-            test_value = value[-1]
+            test_value = value[-1] or ''
         test_value = test_value.strip('%')
         for svalue, test in field['selection']:
             if test.lower().startswith(test_value.lower()):
@@ -695,6 +802,7 @@ def test_complete_selection():
             ('', ['male', 'female']),
             (None, ['male', 'female']),
             (['male', 'f'], [['male', 'female']]),
+            (['male', None], [['male', 'male'], ['male', 'female']]),
             ):
         assert list(complete_value(field, value)) == result
 
@@ -711,6 +819,25 @@ def test_complete_selection():
         assert list(complete_value(field_with_empty, value)) == result
 
 
+def test_complete_reference():
+    field = {
+        'type': 'reference',
+        'selection': [
+            ('spam', 'Spam'),
+            ('ham', 'Ham'),
+            ('', ''),
+            ],
+        }
+    for value, result in (
+            ('s', ['%spam%']),
+            ('test', []),
+            ('', ['%spam%', '%ham%', '%']),
+            (None, ['%spam%', '%ham%', '%']),
+            (['spam', 'h'], [['spam', 'ham']]),
+            ):
+        assert list(complete_value(field, value)) == result
+
+
 def parenthesize(tokens):
     "Nest tokens according to parenthesis"
     for token in tokens:
@@ -741,9 +868,12 @@ def test_parenthesize():
 def operatorize(tokens, operator='or'):
     "Convert operators"
     test = (operator, (operator,))
-    cur = next(tokens)
-    while cur in test:
+    try:
         cur = next(tokens)
+        while cur in test:
+            cur = next(tokens)
+    except StopIteration:
+        return
     if isgenerator(cur):
         cur = operatorize(cur, operator)
     nex = None
@@ -778,6 +908,8 @@ def test_operatorize():
     a = ('a', 'a', 'a')
     b = ('b', 'b', 'b')
     c = ('c', 'c', 'c')
+    null_ = ('d', None, 'x')
+    double_null_ = ('e', None, None)
     for value, result in (
             (['a'], ['a']),
             (['a', 'or', 'b'], [['OR', 'a', 'b']]),
@@ -803,6 +935,8 @@ def test_operatorize():
             (['a', iter(['b', 'or', 'c'])], ['a', [['OR', 'b', 'c']]]),
             ([a, iter([b, ('or',), c])], [a, [['OR', b, c]]]),
             (['a', iter(['b', 'or'])], ['a', [['OR', 'b']]]),
+            ([null_], [null_]),
+            ([null_, 'or', double_null_], [['OR', null_, double_null_]]),
             ):
         assert rlist(operatorize(iter(value))) == result
 
@@ -810,13 +944,14 @@ def test_operatorize():
 class DomainParser(object):
     "A parser for domain"
 
-    def __init__(self, fields):
+    def __init__(self, fields, context=None):
         self.fields = OrderedDict((name, f)
-            for name, f in fields.iteritems()
+            for name, f in fields.items()
             if f.get('searchable', True))
         self.strings = dict((f['string'].lower(), f)
-            for f in fields.itervalues()
+            for f in fields.values()
             if f.get('searchable', True))
+        self.context = context
 
     def parse(self, input_):
         "Return domain for the input string"
@@ -829,8 +964,8 @@ class DomainParser(object):
             tokens = operatorize(tokens, 'and')
             tokens = self.parse_clause(tokens)
             return simplify(rlist(tokens))
-        except ValueError, exception:
-            if exception.message == 'No closing quotation':
+        except ValueError as exception:
+            if str(exception) == 'No closing quotation':
                 return self.parse(input_ + '"')
 
     def stringable(self, domain):
@@ -842,7 +977,24 @@ class DomainParser(object):
                         or isinstance(clause[0], (list, tuple)))
                     and all(isinstance(c, (list, tuple)) for c in clause[1:])):
                 return self.stringable(clause)
-            if clause[0] in self.fields or clause[0] == 'rec_name':
+            name, _, value = clause[:3]
+            if name.endswith('.rec_name'):
+                name = name[:-len('.rec_name')]
+            if name in self.fields:
+                field = self.fields[name]
+                if field['type'] in {
+                        'many2one', 'one2one', 'one2many', 'many2many'}:
+                    if field['type'] == 'many2one':
+                        types = (str, type(None))
+                    else:
+                        types = str
+                    if isinstance(value, (list, tuple)):
+                        return all(isinstance(v, types) for v in value)
+                    else:
+                        return isinstance(value, types)
+                else:
+                    return True
+            elif name == 'rec_name':
                 return True
             return False
 
@@ -858,7 +1010,7 @@ class DomainParser(object):
         def string_(clause):
             if not clause:
                 return ''
-            if (not isinstance(clause[0], basestring)
+            if (not isinstance(clause[0], str)
                     or clause[0] in ('AND', 'OR')):
                 return '(%s)' % self.string(clause)
             name, operator, value = clause[:3]
@@ -902,7 +1054,7 @@ class DomainParser(object):
                     operator = '!'
                 else:
                     operator = ''
-            formatted_value = format_value(field, value, target)
+            formatted_value = format_value(field, value, target, self.context)
             if (operator in OPERATORS and
                     field['type'] in ('char', 'text', 'selection')
                     and value == ''):
@@ -923,7 +1075,7 @@ class DomainParser(object):
         "Return completion for the input string"
         domain = self.parse(input_)
         closing = 0
-        for i in xrange(1, len(input_)):
+        for i in range(1, len(input_)):
             if input_[-i] not in (')', ' '):
                 break
             if input_[-i] == ')':
@@ -946,7 +1098,7 @@ class DomainParser(object):
                 return
             if len(input_) >= 2 and input_[-2] == ':':
                 return
-        for field in self.strings.itervalues():
+        for field in self.strings.values():
             operator = default_operator(field)
             value = ''
             if 'ilike' in operator:
@@ -980,7 +1132,7 @@ class DomainParser(object):
             name = ''
         if (name.lower() not in self.strings
                 and name not in self.fields):
-            for field in self.strings.itervalues():
+            for field in self.strings.values():
                 if field['string'].lower().startswith(name.lower()):
                     operator = default_operator(field)
                     value = ''
@@ -1011,7 +1163,7 @@ class DomainParser(object):
             except ValueError:
                 for part in parts:
                     yield (part,)
-                raise StopIteration
+                return
             for j in range(i):
                 name = ' '.join(parts[j:i])
                 if name.lower() in self.strings:
@@ -1101,20 +1253,21 @@ class DomainParser(object):
                             'datetime', 'date', 'time'):
                         if value and '..' in value:
                             lvalue, rvalue = value.split('..', 1)
-                            lvalue = convert_value(field, lvalue)
-                            rvalue = convert_value(field, rvalue)
+                            lvalue = convert_value(field, lvalue, self.context)
+                            rvalue = convert_value(field, rvalue, self.context)
                             yield iter([
                                     (field_name, '>=', lvalue),
-                                    (field_name, '<', rvalue),
+                                    (field_name, '<=', rvalue),
                                     ])
                             continue
                     if isinstance(value, list):
-                        value = [convert_value(field, v) for v in value]
+                        value = [convert_value(field, v, self.context)
+                            for v in value]
                         if field['type'] in ('many2one', 'one2many',
                                 'many2many', 'one2one'):
                             field_name += '.rec_name'
                     else:
-                        value = convert_value(field, value)
+                        value = convert_value(field, value, self.context)
                     if 'like' in operator:
                         value = likify(value)
                     if target:
@@ -1129,6 +1282,14 @@ def test_stringable():
                 'string': 'Name',
                 'type': 'char',
                 },
+            'relation': {
+                'string': 'Relation',
+                'type': 'many2one',
+                },
+            'relations': {
+                'string': 'Relations',
+                'type': 'many2many',
+                },
             })
     valid = ('name', '=', 'Doe')
     invalid = ('surname', '=', 'John')
@@ -1138,6 +1299,13 @@ def test_stringable():
     assert not dom.stringable(['AND', valid, invalid])
     assert dom.stringable([[valid]])
     assert not dom.stringable([[valid], [invalid]])
+    assert dom.stringable([('relation', '=', None)])
+    assert dom.stringable([('relation', '=', "Foo")])
+    assert dom.stringable([('relation.rec_name', '=', "Foo")])
+    assert not dom.stringable([('relation', '=', 1)])
+    assert dom.stringable([('relations', '=', "Foo")])
+    assert dom.stringable([('relations', 'in', ["Foo"])])
+    assert not dom.stringable([('relations', 'in', [42])])
 
 
 def test_string():
@@ -1237,81 +1405,81 @@ def test_group():
                 'string': '(Sur)Name',
                 },
             })
-    assert rlist(dom.group(udlex(u'Name: Doe'))) == [('Name', None, 'Doe')]
-    assert rlist(dom.group(udlex(u'"(Sur)Name": Doe'))) == [
+    assert rlist(dom.group(udlex('Name: Doe'))) == [('Name', None, 'Doe')]
+    assert rlist(dom.group(udlex('"(Sur)Name": Doe'))) == [
         ('(Sur)Name', None, 'Doe'),
         ]
-    assert rlist(dom.group(udlex(u'Name: Doe Name: John'))) == [
+    assert rlist(dom.group(udlex('Name: Doe Name: John'))) == [
         ('Name', None, 'Doe'),
         ('Name', None, 'John')]
-    assert rlist(dom.group(udlex(u'Name: Name: John'))) == [
+    assert rlist(dom.group(udlex('Name: Name: John'))) == [
         ('Name', None, None),
         ('Name', None, 'John')]
-    assert rlist(dom.group(udlex(u'First Name: John'))) == [
+    assert rlist(dom.group(udlex('First Name: John'))) == [
         ('First Name', None, 'John'),
         ]
-    assert rlist(dom.group(udlex(u'Name: Doe First Name: John'))) == [
+    assert rlist(dom.group(udlex('Name: Doe First Name: John'))) == [
         ('Name', None, 'Doe'),
         ('First Name', None, 'John'),
         ]
-    assert rlist(dom.group(udlex(u'First Name: John Name: Doe'))) == [
+    assert rlist(dom.group(udlex('First Name: John Name: Doe'))) == [
         ('First Name', None, 'John'),
         ('Name', None, 'Doe'),
         ]
-    assert rlist(dom.group(udlex(u'First Name: John First Name: Jane'))) == [
+    assert rlist(dom.group(udlex('First Name: John First Name: Jane'))) == [
         ('First Name', None, 'John'),
         ('First Name', None, 'Jane'),
         ]
-    assert rlist(dom.group(udlex(u'Name: John Doe'))) == [
+    assert rlist(dom.group(udlex('Name: John Doe'))) == [
         ('Name', None, 'John'),
         ('Doe',),
         ]
-    assert rlist(dom.group(udlex(u'Name: "John Doe"'))) == [
+    assert rlist(dom.group(udlex('Name: "John Doe"'))) == [
         ('Name', None, 'John Doe'),
         ]
-    assert rlist(dom.group(udlex(u'Name: =Doe'))) == [('Name', '=', 'Doe')]
-    assert rlist(dom.group(udlex(u'Name: =Doe Name: >John'))) == [
+    assert rlist(dom.group(udlex('Name: =Doe'))) == [('Name', '=', 'Doe')]
+    assert rlist(dom.group(udlex('Name: =Doe Name: >John'))) == [
         ('Name', '=', 'Doe'),
         ('Name', '>', 'John'),
         ]
-    assert rlist(dom.group(udlex(u'First Name: =John First Name: =Jane'))) == [
+    assert rlist(dom.group(udlex('First Name: =John First Name: =Jane'))) == [
         ('First Name', '=', 'John'),
         ('First Name', '=', 'Jane'),
         ]
-    assert rlist(dom.group(udlex(u'Name: John;Jane'))) == [
+    assert rlist(dom.group(udlex('Name: John;Jane'))) == [
         ('Name', None, ['John', 'Jane'])
         ]
-    assert rlist(dom.group(udlex(u'Name: John;'))) == [
+    assert rlist(dom.group(udlex('Name: John;'))) == [
         ('Name', None, ['John'])
         ]
-    assert rlist(dom.group(udlex(u'Name: John;Jane Name: Doe'))) == [
+    assert rlist(dom.group(udlex('Name: John;Jane Name: Doe'))) == [
         ('Name', None, ['John', 'Jane']),
         ('Name', None, 'Doe'),
         ]
-    assert rlist(dom.group(udlex(u'Name: John; Name: Doe'))) == [
+    assert rlist(dom.group(udlex('Name: John; Name: Doe'))) == [
         ('Name', None, ['John']),
         ('Name', None, 'Doe'),
         ]
-    assert rlist(dom.group(udlex(u'Name:'))) == [
+    assert rlist(dom.group(udlex('Name:'))) == [
         ('Name', None, None),
         ]
-    assert rlist(dom.group(udlex(u'Name: ='))) == [
+    assert rlist(dom.group(udlex('Name: ='))) == [
         ('Name', '=', None),
         ]
-    assert rlist(dom.group(udlex(u'Name: =""'))) == [
+    assert rlist(dom.group(udlex('Name: =""'))) == [
         ('Name', '=', ''),
         ]
-    assert rlist(dom.group(udlex(u'Name: = ""'))) == [
+    assert rlist(dom.group(udlex('Name: = ""'))) == [
         ('Name', '=', ''),
         ]
-    assert rlist(dom.group(udlex(u'Name: = Name: Doe'))) == [
+    assert rlist(dom.group(udlex('Name: = Name: Doe'))) == [
         ('Name', '=', None),
         ('Name', None, 'Doe'),
         ]
-    assert rlist(dom.group(udlex(u'Name: \\"foo\\"'))) == [
+    assert rlist(dom.group(udlex('Name: \\"foo\\"'))) == [
         ('Name', None, '"foo"'),
         ]
-    assert rlist(dom.group(udlex(u'Name: "" <'))) == [
+    assert rlist(dom.group(udlex('Name: "" <'))) == [
         ('Name', '', '<'),
         ]
 
@@ -1387,7 +1555,7 @@ def test_parse_clause():
         ]
     assert rlist(dom.parse_clause([('Integer', None, '3..5')])) == [[
             ('integer', '>=', 3),
-            ('integer', '<', 5),
+            ('integer', '<=', 5),
             ]]
     assert rlist(dom.parse_clause([('Reference', None, 'foo')])) == [
         ('reference', 'ilike', '%foo%'),
@@ -1401,12 +1569,19 @@ def test_parse_clause():
     assert rlist(dom.parse_clause([('Reference', None, ['foo', 'bar'])])) == [
         ('reference', 'in', ['foo', 'bar']),
         ]
+    assert rlist(dom.parse_clause(['OR',
+                ('Name', None, 'John'), ('Name', None, 'Jane')])) == ['OR',
+        ('name', 'ilike', '%John%'),
+        ('name', 'ilike', '%Jane%'),
+        ]
     assert rlist(dom.parse_clause([('Many2One', None, 'John')])) == [
         ('many2one', 'ilike', '%John%'),
         ]
     assert rlist(dom.parse_clause([('Many2One', None, ['John', 'Jane'])])) == [
         ('many2one.rec_name', 'in', ['John', 'Jane']),
         ]
+    assert rlist(dom.parse_clause(iter([iter([['John']])]))) == [
+        [('rec_name', 'ilike', '%John%')]]
 
 
 def test_completion():
@@ -1417,8 +1592,17 @@ def test_completion():
                 'type': 'char',
                 },
             })
-    assert list(dom.completion(u'Nam')) == ['Name: ']
-    assert list(dom.completion(u'Name:')) == ['Name: ']
-    assert list(dom.completion(u'Name: foo')) == []
-    assert list(dom.completion(u'Name: !=')) == []
-    assert list(dom.completion(u'Name: !=foo')) == []
+    assert list(dom.completion('Nam')) == ['Name: ']
+    assert list(dom.completion('Name:')) == ['Name: ']
+    assert list(dom.completion('Name: foo')) == []
+    assert list(dom.completion('Name: !=')) == []
+    assert list(dom.completion('Name: !=foo')) == []
+    assert list(dom.completion('')) == ['Name: ']
+    assert list(dom.completion(' ')) == ['', 'Name: ']
+
+
+if __name__ == '__main__':
+    for name in list(globals()):
+        if name.startswith('test_'):
+            func = globals()[name]
+            func()
