@@ -1,15 +1,17 @@
-#This file is part of Tryton.  The COPYRIGHT file at the top level of
-#this repository contains the full copyright notices and license terms.
+# This file is part of Tryton.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
 "Preference"
 import gettext
 import gtk
 import copy
-from tryton.gui.window.view_form.screen import Screen
-from tryton.config import TRYTON_ICON
-import tryton.common as common
-from tryton.common import RPCExecute, RPCException
-from tryton.gui.window.nomodal import NoModal
+
 import tryton.rpc as rpc
+from tryton.common import RPCExecute, RPCException, IconFactory
+from tryton.common.underline import set_underline
+from tryton.config import TRYTON_ICON
+from tryton.gui import Main
+from tryton.gui.window.nomodal import NoModal
+from tryton.gui.window.view_form.screen import Screen
 
 _ = gettext.gettext
 
@@ -22,16 +24,23 @@ class Preference(NoModal):
         self.callback = callback
         self.win = gtk.Dialog(_('Preferences'), self.parent,
             gtk.DIALOG_DESTROY_WITH_PARENT)
+        Main().add_window(self.win)
         self.win.set_position(gtk.WIN_POS_CENTER_ON_PARENT)
-        self.win.set_has_separator(False)
         self.win.set_icon(TRYTON_ICON)
 
         self.accel_group = gtk.AccelGroup()
         self.win.add_accel_group(self.accel_group)
 
-        self.but_cancel = self.win.add_button(gtk.STOCK_CANCEL,
-                gtk.RESPONSE_CANCEL)
-        self.but_ok = self.win.add_button(gtk.STOCK_OK, gtk.RESPONSE_OK)
+        self.but_cancel = self.win.add_button(
+            set_underline(_("Cancel")), gtk.RESPONSE_CANCEL)
+        self.but_cancel.set_image(IconFactory.get_image(
+                'tryton-cancel', gtk.ICON_SIZE_BUTTON))
+        self.but_cancel.set_always_show_image(True)
+        self.but_ok = self.win.add_button(
+            set_underline(_("OK")), gtk.RESPONSE_OK)
+        self.but_ok.set_image(IconFactory.get_image(
+                'tryton-ok', gtk.ICON_SIZE_BUTTON))
+        self.but_ok.set_always_show_image(True)
         self.but_ok.add_accelerator('clicked', self.accel_group,
                 gtk.keysyms.Return, gtk.gdk.CONTROL_MASK, gtk.ACCEL_VISIBLE)
 
@@ -65,6 +74,7 @@ class Preference(NoModal):
             self.win.destroy()
             self.win = None
             return
+        self.screen.current_record.cancel()
         self.screen.current_record.set(preferences)
         self.screen.current_record.id = rpc._USER
         self.screen.current_record.validate(softvalidation=True)
@@ -75,7 +85,7 @@ class Preference(NoModal):
         self.win.set_title(_('Preference'))
 
         width, height = self.parent.get_size()
-        self.win.set_default_size(int(width * 0.9), int(height * 0.9))
+        self.win.set_default_size(width, height)
 
         self.register()
         self.win.show()
@@ -84,16 +94,8 @@ class Preference(NoModal):
         if response_id == gtk.RESPONSE_OK:
             if self.screen.current_record.validate():
                 vals = copy.copy(self.screen.get())
-                if 'password' in vals:
-                    password = common.ask(_('Current Password:'),
-                        visibility=False)
-                    if not password:
-                        return
-                else:
-                    password = False
                 try:
-                    RPCExecute('model', 'res.user', 'set_preferences',
-                        vals, password)
+                    RPCExecute('model', 'res.user', 'set_preferences', vals)
                 except RPCException:
                     return
         self.parent.present()
