@@ -1,20 +1,22 @@
-#This file is part of Tryton.  The COPYRIGHT file at the top level of
-#this repository contains the full copyright notices and license terms.
+# This file is part of Tryton.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
 import os
 from itertools import chain
 import tempfile
 import locale
-from tryton.common import datetime_strftime, \
+from tryton.common import \
         domain_inversion, eval_domain, localize_domain, \
-        merge, inverse_leaf, concat, simplify, EvalEnvironment
+        merge, inverse_leaf, filter_leaf, prepare_reference_domain, \
+        extract_reference_models, concat, simplify, unique_value, \
+        EvalEnvironment
 import tryton.common as common
-import time
 import datetime
 import decimal
 from decimal import Decimal
 import math
-from tryton.translate import date_format
 from tryton.common import RPCExecute, RPCException
+from tryton.pyson import PYSONDecoder
+from tryton.config import CONFIG
 
 
 class Field(object):
@@ -33,15 +35,17 @@ class Field(object):
     def __init__(self, attrs):
         self.attrs = attrs
         self.name = attrs['name']
+        self.views = set()
 
     def sig_changed(self, record):
-        if self.attrs.get('on_change', False):
-            record.on_change(self.name, self.attrs['on_change'])
-        record.on_change_with(self.name)
+        record.on_change([self.name])
+        record.on_change_with([self.name])
         record.autocomplete_with(self.name)
+        record.set_field_context()
 
-    def domains_get(self, record):
-        screen_domain = domain_inversion(record.group.domain4inversion,
+    def domains_get(self, record, pre_validate=None):
+        screen_domain = domain_inversion(
+            [record.group.domain4inversion, pre_validate or []],
             self.name, EvalEnvironment(record))
         if isinstance(screen_domain, bool) and not screen_domain:
             screen_domain = [('id', '=', None)]
@@ -54,16 +58,26 @@ class Field(object):
         screen_domain, attr_domain = self.domains_get(record)
         return concat(localize_domain(screen_domain), attr_domain)
 
-    def validation_domains(self, record):
-        return concat(*self.domains_get(record))
+    def validation_domains(self, record, pre_validate=None):
+        return concat(*self.domains_get(record, pre_validate))
 
-    def context_get(self, record):
-        context = record.context_get().copy()
-        if record.parent:
-            context.update(record.parent.context_get())
+    def get_context(self, record, record_context=None, local=False):
+        if record_context is not None:
+            context = record_context.copy()
+        else:
+            context = record.get_context(local=local)
         context.update(record.expr_eval(self.attrs.get('context', {})))
         return context
 
+    def get_search_context(self, record):
+        context = self.get_context(record)
+        context.update(record.expr_eval(self.attrs.get('search_context', {})))
+        return context
+
+    def get_search_order(self, record):
+        order = record.expr_eval(self.attrs.get('search_order', None))
+        return order
+
     def _is_empty(self, record):
         return not self.get_eval(record)
 
@@ -75,27 +89,27 @@ class Field(object):
                 return False
         return True
 
-    def validate(self, record, softvalidation=False):
+    def validate(self, record, softvalidation=False, pre_validate=None):
         if self.attrs.get('readonly'):
             return True
-        res = True
+        invalid = False
         self.get_state_attrs(record)['domain_readonly'] = False
-        domain = simplify(self.validation_domains(record))
+        domain = simplify(self.validation_domains(record, pre_validate))
         if not softvalidation:
-            res = res and self.check_required(record)
+            if not self.check_required(record):
+                invalid = 'required'
         if isinstance(domain, bool):
-            res = res and domain
+            if not domain:
+                invalid = 'domain'
         elif domain == [('id', '=', None)]:
-            res = False
+            invalid = 'domain'
         else:
-            if (isinstance(domain, list)
-                    and len(domain) == 1
-                    and domain[0][1] == '='):
+            unique, leftpart, value = unique_value(domain)
+            if unique:
                 # If the inverted domain is so constraint that only one value
                 # is possible we should use it. But we must also pay attention
                 # to the fact that the original domain might be a 'OR' domain
                 # and thus not preventing the modification of fields.
-                leftpart, _, value = domain[0][:3]
                 if value is False:
                     # XXX to remove once server domains are fixed
                     value = None
@@ -113,13 +127,14 @@ class Field(object):
                             constraintfields.add(leaf[0])
                     if localpart != 'id' or recordpart not in constraintfields:
                         setdefault = False
-                if setdefault:
+                if setdefault and not pre_validate:
                     self.set_client(record, value)
                     self.get_state_attrs(record)['domain_readonly'] = (
                         domain_readonly)
-            res = res and eval_domain(domain, EvalEnvironment(record))
-        self.get_state_attrs(record)['valid'] = res
-        return res
+            if not eval_domain(domain, EvalEnvironment(record)):
+                invalid = domain
+        self.get_state_attrs(record)['invalid'] = invalid
+        return not invalid
 
     def set(self, record, value):
         record.value[self.name] = value
@@ -156,7 +171,7 @@ class Field(object):
 
     def set_on_change(self, record, value):
         record.modified_fields.setdefault(self.name)
-        return self.set(record, value)
+        self.set(record, value)
 
     def state_set(self, record, states=('readonly', 'required', 'invisible')):
         state_changes = record.expr_eval(self.attrs.get('states', {}))
@@ -202,12 +217,20 @@ class DateTimeField(Field):
     _default = None
 
     def set_client(self, record, value, force_change=False):
-        if isinstance(value, basestring):
-            try:
-                value = datetime.datetime(*time.strptime(value,
-                        date_format() + ' ' + self.time_format(record))[:6])
-            except ValueError:
-                value = self._default
+        if isinstance(value, datetime.time):
+            current_value = self.get_client(record)
+            if current_value:
+                value = datetime.datetime.combine(
+                    current_value.date(), value)
+            else:
+                value = None
+        elif value and not isinstance(value, datetime.datetime):
+            current_value = self.get_client(record)
+            if current_value:
+                time = current_value.time()
+            else:
+                time = datetime.time()
+            value = datetime.datetime.combine(value, time)
         if value:
             value = common.untimezoned_date(value)
         super(DateTimeField, self).set_client(record, value,
@@ -216,10 +239,11 @@ class DateTimeField(Field):
     def get_client(self, record):
         value = super(DateTimeField, self).get_client(record)
         if value:
-            value = common.timezoned_date(value)
-            return datetime_strftime(value, date_format() + ' ' +
-                self.time_format(record))
-        return ''
+            return common.timezoned_date(value)
+
+    def date_format(self, record):
+        context = self.get_context(record)
+        return common.date_format(context.get('date_format'))
 
     def time_format(self, record):
         return record.expr_eval(self.attrs['format'])
@@ -230,23 +254,15 @@ class DateField(Field):
     _default = None
 
     def set_client(self, record, value, force_change=False):
-        if isinstance(value, basestring):
-            try:
-                value = datetime.date(*time.strptime(value,
-                        date_format())[:3])
-            except ValueError:
-                value = self._default
-        elif isinstance(value, datetime.datetime):
+        if isinstance(value, datetime.datetime):
             assert(value.time() == datetime.time())
             value = value.date()
         super(DateField, self).set_client(record, value,
             force_change=force_change)
 
-    def get_client(self, record):
-        value = super(DateField, self).get_client(record)
-        if value:
-            return datetime_strftime(value, date_format())
-        return ''
+    def date_format(self, record):
+        context = self.get_context(record)
+        return common.date_format(context.get('date_format'))
 
 
 class TimeField(Field):
@@ -257,30 +273,38 @@ class TimeField(Field):
         return self.get(record) is None
 
     def set_client(self, record, value, force_change=False):
-        if isinstance(value, basestring):
-            try:
-                value = datetime.time(*time.strptime(value,
-                        self.time_format(record))[3:6])
-            except ValueError:
-                value = self._default
-        elif isinstance(value, datetime.datetime):
+        if isinstance(value, datetime.datetime):
             value = value.time()
         super(TimeField, self).set_client(record, value,
             force_change=force_change)
 
-    def get_client(self, record):
-        value = super(TimeField, self).get_client(record)
-        if value is not None:
-            return value.strftime(self.time_format(record))
-        return ''
-
     def time_format(self, record):
         return record.expr_eval(self.attrs['format'])
 
 
+class TimeDeltaField(Field):
+
+    _default = None
+
+    def _is_empty(self, record):
+        return self.get(record) is None
+
+    def converter(self, group):
+        return group.context.get(self.attrs.get('converter'))
+
+    def set_client(self, record, value, force_change=False):
+        if isinstance(value, str):
+            value = common.timedelta.parse(value, self.converter(record.group))
+        super(TimeDeltaField, self).set_client(
+            record, value, force_change=force_change)
+
+    def get_client(self, record):
+        value = super(TimeDeltaField, self).get_client(record)
+        return common.timedelta.format(value, self.converter(record.group))
+
+
 class FloatField(Field):
     _default = None
-    default_digits = (16, 2)
 
     def _is_empty(self, record):
         return self.get(record) is None
@@ -289,10 +313,9 @@ class FloatField(Field):
         return record.value.get(self.name, self._default)
 
     def digits(self, record, factor=1):
-        digits = tuple(y if x is None else x for x, y in zip(
-                record.expr_eval(
-                    self.attrs.get('digits', self.default_digits)),
-                self.default_digits))
+        digits = record.expr_eval(self.attrs.get('digits'))
+        if not digits or any(d is None for d in digits):
+            return
         shift = int(round(math.log(abs(factor), 10)))
         return (digits[0] + shift, digits[1] - shift)
 
@@ -303,18 +326,29 @@ class FloatField(Field):
             return self._default
 
     def set_client(self, record, value, force_change=False, factor=1):
-        if isinstance(value, basestring):
+        if isinstance(value, str):
             value = self.convert(value)
         if value is not None:
-            value /= factor
+            # Keep the same type
+            if isinstance(value, int):
+                value //= factor
+            else:
+                value /= factor
         super(FloatField, self).set_client(record, value,
             force_change=force_change)
 
     def get_client(self, record, factor=1):
         value = record.value.get(self.name)
         if value is not None:
-            digit = self.digits(record, factor=factor)[1]
-            return locale.format('%.*f', (digit, value * factor), True)
+            digits = self.digits(record, factor=factor)
+            if digits:
+                p = int(digits[1])
+            else:
+                d = value * factor
+                if not isinstance(d, Decimal):
+                    d = Decimal(repr(d))
+                p = -int(d.as_tuple().exponent)
+            return locale.format('%.*f', (p, value * factor), True)
         else:
             return ''
 
@@ -337,7 +371,6 @@ class NumericField(FloatField):
 
 
 class IntegerField(FloatField):
-    default_digits = (16, 0)
 
     def convert(self, value):
         try:
@@ -376,12 +409,8 @@ class M2OField(Field):
 
     _default = None
 
-    def get(self, record):
-        value = record.value.get(self.name)
-        if (record.parent_name == self.name
-                and self.attrs['relation'] == record.group.parent.model_name):
-            value = record.parent.id if record.parent else None
-        return value
+    def _is_empty(self, record):
+        return self.get(record) is None
 
     def get_client(self, record):
         rec_name = record.value.get(self.name + '.rec_name')
@@ -404,40 +433,26 @@ class M2OField(Field):
 
     def set(self, record, value):
         rec_name = record.value.get(self.name + '.rec_name') or ''
-        if value is False:
-            value = None
-        if (record.parent_name == self.name
-                and self.attrs['relation'] == record.group.parent.model_name):
-            if record.parent:
-                value = record.parent.id
-                if 'rec_name' in record.parent.value:
-                    rec_name = record.parent.value['rec_name'] or ''
-            else:
-                value = None
-        if not rec_name and value >= 0:
+        if not rec_name and value is not None and value >= 0:
             try:
                 result, = RPCExecute('model', self.attrs['relation'], 'read',
                     [value], ['rec_name'])
             except RPCException:
-                return False
+                return
             rec_name = result['rec_name'] or ''
         record.value[self.name + '.rec_name'] = rec_name
         record.value[self.name] = value
-        if (record.parent_name == self.name
-                and self.attrs['relation'] == record.group.parent.model_name):
-            if record.parent:
-                if 'rec_name' not in record.parent.value:
-                    record.parent.value['rec_name'] = rec_name
 
-    def context_get(self, record):
-        context = super(M2OField, self).context_get(record)
+    def get_context(self, record, record_context=None, local=False):
+        context = super(M2OField, self).get_context(
+            record, record_context=record_context, local=local)
         if self.attrs.get('datetime_field'):
             context['_datetime'] = record.get_eval(
                 )[self.attrs.get('datetime_field')]
         return context
 
-    def validation_domains(self, record):
-        screen_domain, attr_domain = self.domains_get(record)
+    def validation_domains(self, record, pre_validate=None):
+        screen_domain, attr_domain = self.domains_get(record, pre_validate)
         return screen_domain
 
     def domain_get(self, record):
@@ -445,13 +460,11 @@ class M2OField(Field):
         return concat(localize_domain(inverse_leaf(screen_domain), self.name),
             attr_domain)
 
-    def get_state_attrs(self, record):
-        result = super(M2OField, self).get_state_attrs(record)
-        if (record.parent_name == self.name
-                and self.attrs['relation'] == record.group.parent.model_name):
-            result = result.copy()
-            result['readonly'] = True
-        return result
+    def get_on_change_value(self, record):
+        if record.parent_name == self.name and record.parent:
+            return record.parent.get_on_change_value(
+                skip={record.group.child_name})
+        return super(M2OField, self).get_on_change_value(record)
 
 
 class O2OField(M2OField):
@@ -467,12 +480,6 @@ class O2MField(Field):
 
     def __init__(self, attrs):
         super(O2MField, self).__init__(attrs)
-        self.in_on_change = False
-        self.context = {}
-
-    def sig_changed(self, record):
-        if not self.in_on_change:
-            return super(O2MField, self).sig_changed(record)
 
     def _group_changed(self, group, record):
         if not record.parent:
@@ -500,14 +507,15 @@ class O2MField(Field):
     def _set_default_value(self, record, fields=None):
         if record.value.get(self.name) is not None:
             return
-        from group import Group
+        from .group import Group
         parent_name = self.attrs.get('relation_field', '')
         fields = fields or {}
+        context = record.expr_eval(self.attrs.get('context', {}))
         group = Group(self.attrs['relation'], fields,
                 parent=record,
                 parent_name=parent_name,
                 child_name=self.name,
-                context=self.context,
+                context=context,
                 parent_datetime_field=self.attrs.get('datetime_field'))
         if not fields and record.model_name == self.attrs['relation']:
             group.fields = record.group.fields
@@ -539,11 +547,12 @@ class O2MField(Field):
             if record2 in record_removed or record2 in record_deleted:
                 continue
             if record2.id >= 0:
-                values = record2.get()
-                values.pop(parent_name, None)
-                if record2.modified and values:
-                    to_write.extend(([record2.id], values))
-                to_add.append(record2.id)
+                if record2.modified:
+                    values = record2.get()
+                    values.pop(parent_name, None)
+                    if values:
+                        to_write.extend(([record2.id], values))
+                    to_add.append(record2.id)
             else:
                 values = record2.get()
                 values.pop(parent_name, None)
@@ -564,9 +573,10 @@ class O2MField(Field):
         if record.value.get(self.name) is None:
             return {}
         result = {}
-        for record2 in (record.value[self.name]
-                + record.value[self.name].record_removed
-                + record.value[self.name].record_deleted):
+        record_modified = (r for r in record.value[self.name] if r.modified)
+        for record2 in chain(record_modified,
+                record.value[self.name].record_removed,
+                record.value[self.name].record_deleted):
             result.update(record2.get_timestamp())
         return result
 
@@ -585,23 +595,24 @@ class O2MField(Field):
         for record2 in record.value[self.name]:
             if not (record2.deleted or record2.removed):
                 result.append(
-                    record2.get_on_change_value())
+                    record2.get_on_change_value(
+                        skip={self.attrs.get('relation_field', '')}))
         return result
 
-    def _set_value(self, record, value, default=False):
+    def _set_value(self, record, value, default=False, modified=False):
         self._set_default_value(record)
         group = record.value[self.name]
         if value is None:
             value = []
-        if not value or isinstance(value[0], (int, long)):
+        if not value or isinstance(value[0], int):
             mode = 'list ids'
         else:
             mode = 'list values'
 
         if mode == 'list values':
-            context = self.context_get(record)
+            context = self.get_context(record)
             field_names = set(f for v in value for f in v
-                if f not in group.fields)
+                if f not in group.fields and '.' not in f)
             if field_names:
                 try:
                     fields = RPCExecute('model', self.attrs['relation'],
@@ -614,16 +625,19 @@ class O2MField(Field):
             for old_record in group:
                 if old_record.id not in value:
                     group.remove(old_record, remove=True, signal=False)
-            group.load(value)
+            group.load(value, modified=modified)
         else:
             for vals in value:
                 new_record = record.value[self.name].new(default=False)
                 if default:
-                    new_record.set_default(vals)
-                    group.add(new_record)
+                    # Don't validate as parent will validate
+                    new_record.set_default(vals, signal=False, validate=False)
+                    group.add(new_record, signal=False)
                 else:
-                    new_record.set(vals)
+                    new_record.set(vals, signal=False)
                     group.append(new_record)
+            # Trigger signal only once with the last record
+            new_record.signal('record-changed')
 
     def set(self, record, value, _default=False):
         group = record.value.get(self.name)
@@ -637,7 +651,7 @@ class O2MField(Field):
             fields = record.group.fields
         if fields:
             fields = dict((fname, field.attrs)
-                for fname, field in fields.iteritems())
+                for fname, field in fields.items())
 
         record.value[self.name] = None
         self._set_default_value(record, fields=fields)
@@ -652,13 +666,14 @@ class O2MField(Field):
         if value is None:
             value = []
         # domain inversion could try to set id as value
-        if isinstance(value, (int, long)):
+        if isinstance(value, int):
             value = [value]
 
         previous_ids = self.get_eval(record)
-        self._set_value(record, value)
         # The order of the ids is not significant
-        if set(previous_ids) != set(value):
+        modified = set(previous_ids) != set(value)
+        self._set_value(record, value, modified=modified)
+        if modified:
             record.modified_fields.setdefault(self.name)
             record.signal('record-modified')
             self.sig_changed(record)
@@ -674,30 +689,30 @@ class O2MField(Field):
         record.modified_fields.setdefault(self.name)
 
     def set_on_change(self, record, value):
+        record.modified_fields.setdefault(self.name)
+        self._set_default_value(record)
         if isinstance(value, (list, tuple)):
-            self._set_value(record, value)
-            record.modified_fields.setdefault(self.name)
-            record.signal('record-modified')
-            return True
+            self._set_value(record, value, modified=True)
+            return
 
         if value and (value.get('add') or value.get('update')):
-            context = self.context_get(record)
+            context = self.get_context(record)
             fields = record.value[self.name].fields
             values = chain(value.get('update', []),
                 (d for _, d in value.get('add', [])))
             field_names = set(f for v in values
-                for f in v if f not in fields and f != 'id')
+                for f in v if f not in fields and f != 'id' and '.' not in f)
             if field_names:
                 try:
                     fields = RPCExecute('model', self.attrs['relation'],
                         'fields_get', list(field_names), context=context)
                 except RPCException:
-                    return False
+                    return
             else:
                 fields = {}
 
         to_remove = []
-        for record2 in record.value.get(self.name, []):
+        for record2 in record.value[self.name]:
             if not record2.id:
                 to_remove.append(record2)
         if value and value.get('remove'):
@@ -710,10 +725,10 @@ class O2MField(Field):
                 force_remove=False)
 
         if value and (value.get('add') or value.get('update', [])):
-            record.value[self.name].add_fields(fields, signal=False)
+            record.value[self.name].add_fields(fields)
             for index, vals in value.get('add', []):
                 new_record = record.value[self.name].new(default=False)
-                record.value[self.name].add(new_record, index)
+                record.value[self.name].add(new_record, index, signal=False)
                 new_record.set_on_change(vals)
 
             for vals in value.get('update', []):
@@ -722,34 +737,37 @@ class O2MField(Field):
                 record2 = record.value[self.name].get(vals['id'])
                 if record2 is not None:
                     record2.set_on_change(vals)
-        return True
 
-    def validation_domains(self, record):
-        screen_domain, attr_domain = self.domains_get(record)
+    def validation_domains(self, record, pre_validate=None):
+        screen_domain, attr_domain = self.domains_get(record, pre_validate)
         return screen_domain
 
-    def validate(self, record, softvalidation=False):
-        if self.attrs.get('readonly'):
-            return True
-        res = True
+    def validate(self, record, softvalidation=False, pre_validate=None):
+        invalid = False
+        ldomain = localize_domain(domain_inversion(
+                record.group.clean4inversion(pre_validate or []), self.name,
+                EvalEnvironment(record)), self.name)
+        if isinstance(ldomain, bool):
+            if ldomain:
+                ldomain = []
+            else:
+                ldomain = [('id', '=', None)]
         for record2 in record.value.get(self.name, []):
-            if not record2.loaded and record2.id >= 0:
+            if not record2.loaded and record2.id >= 0 and not pre_validate:
                 continue
-            if not record2.validate(softvalidation=softvalidation):
-                if not record2.modified:
-                    record.value[self.name].remove(record2)
-                else:
-                    res = False
-        if not super(O2MField, self).validate(record, softvalidation):
-            res = False
-        self.get_state_attrs(record)['valid'] = res
-        return res
+            if not record2.validate(softvalidation=softvalidation,
+                    pre_validate=ldomain):
+                invalid = 'children'
+        test = super(O2MField, self).validate(record, softvalidation,
+            pre_validate)
+        if test and invalid:
+            self.get_state_attrs(record)['invalid'] = invalid
+            return False
+        return test
 
     def state_set(self, record, states=('readonly', 'required', 'invisible')):
         self._set_default_value(record)
         super(O2MField, self).state_set(record, states=states)
-        record.value[self.name].readonly = self.get_state_attrs(record).get(
-            'readonly', False)
 
     def get_removed_ids(self, record):
         return [x.id for x in record.value[self.name].record_removed]
@@ -771,6 +789,13 @@ class ReferenceField(Field):
 
     _default = None
 
+    def _is_empty(self, record):
+        result = super(ReferenceField, self)._is_empty(record)
+        if not result and (record.value[self.name] is None
+                or record.value[self.name][1] < 0):
+            result = True
+        return result
+
     def get_client(self, record):
         if record.value.get(self.name):
             model, _ = record.value[self.name]
@@ -780,21 +805,16 @@ class ReferenceField(Field):
             return None
 
     def get(self, record):
-        if record.parent_name == self.name:
-            if record.parent:
-                return '%s,%s' % (record.group.parent.model_name,
-                    record.parent.id)
-            else:
-                return None
         if (record.value.get(self.name)
                 and record.value[self.name][0]
+                and record.value[self.name][1] is not None
                 and record.value[self.name][1] >= -1):
             return ','.join(map(str, record.value[self.name]))
         return None
 
     def set_client(self, record, value, force_change=False):
         if value:
-            if isinstance(value, basestring):
+            if isinstance(value, str):
                 value = value.split(',')
             ref_model, ref_id = value
             if isinstance(ref_id, (tuple, list)):
@@ -815,16 +835,10 @@ class ReferenceField(Field):
             force_change=force_change)
 
     def set(self, record, value):
-        if record.parent_name == self.name:
-            if record.parent:
-                value = '%s,%s' % (record.group.parent.model_name,
-                    record.parent.id)
-            else:
-                value = None
         if not value:
             record.value[self.name] = self._default
             return
-        if isinstance(value, basestring):
+        if isinstance(value, str):
             ref_model, ref_id = value.split(',')
             if not ref_id:
                 ref_id = None
@@ -836,7 +850,7 @@ class ReferenceField(Field):
         else:
             ref_model, ref_id = value
         rec_name = record.value.get(self.name + '.rec_name') or ''
-        if ref_model and ref_id >= 0:
+        if ref_model and ref_id is not None and ref_id >= 0:
             if not rec_name and ref_id >= 0:
                 try:
                     result, = RPCExecute('model', ref_model, 'read', [ref_id],
@@ -851,6 +865,45 @@ class ReferenceField(Field):
         record.value[self.name] = ref_model, ref_id
         record.value[self.name + '.rec_name'] = rec_name
 
+    def get_context(self, record, record_context=None, local=False):
+        context = super(ReferenceField, self).get_context(
+            record, record_context, local=local)
+        if self.attrs.get('datetime_field'):
+            context['_datetime'] = record.get_eval(
+                )[self.attrs.get('datetime_field')]
+        return context
+
+    def get_on_change_value(self, record):
+        if record.parent_name == self.name and record.parent:
+            return record.parent.model_name, record.parent.get_on_change_value(
+                skip={record.group.child_name})
+        return super(ReferenceField, self).get_on_change_value(record)
+
+    def validation_domains(self, record, pre_validate=None):
+        screen_domain, attr_domain = self.domains_get(record, pre_validate)
+        return screen_domain
+
+    def domain_get(self, record):
+        if record.value.get(self.name):
+            model = record.value[self.name][0]
+        else:
+            model = None
+        screen_domain, attr_domain = self.domains_get(record)
+        screen_domain = prepare_reference_domain(screen_domain, self.name)
+        return concat(localize_domain(
+                filter_leaf(screen_domain, self.name, model),
+                strip_target=True), attr_domain)
+
+    def get_models(self, record):
+        screen_domain, attr_domain = self.domains_get(record)
+        return extract_reference_models(
+            concat(screen_domain, attr_domain), self.name)
+
+
+class _FileCache(object):
+    def __init__(self, path):
+        self.path = path
+
 
 class BinaryField(Field):
 
@@ -858,10 +911,10 @@ class BinaryField(Field):
 
     def get(self, record):
         result = record.value.get(self.name, self._default)
-        if isinstance(result, basestring):
+        if isinstance(result, _FileCache):
             try:
-                with open(result, 'rb') as fp:
-                    result = buffer(fp.read())
+                with open(result.path, 'rb') as fp:
+                    result = fp.read()
             except IOError:
                 result = self.get_data(record)
         return result
@@ -871,9 +924,12 @@ class BinaryField(Field):
 
     def set_client(self, record, value, force_change=False):
         _, filename = tempfile.mkstemp(prefix='tryton_')
+        data = value or b''
+        if isinstance(data, str):
+            data = data.encode('utf-8')
         with open(filename, 'wb') as fp:
-            fp.write(value or '')
-        self.set(record, filename)
+            fp.write(data)
+        self.set(record, _FileCache(filename))
         record.modified_fields.setdefault(self.name)
         record.signal('record-modified')
         self.sig_changed(record)
@@ -882,26 +938,30 @@ class BinaryField(Field):
 
     def get_size(self, record):
         result = record.value.get(self.name) or 0
-        if isinstance(result, basestring):
-            result = os.stat(result).st_size
-        elif isinstance(result, buffer):
+        if isinstance(result, _FileCache):
+            result = os.stat(result.path).st_size
+        elif isinstance(result, (str, bytes)):
             result = len(result)
         return result
 
     def get_data(self, record):
-        if not isinstance(record.value.get(self.name), (basestring, buffer)):
+        if not isinstance(record.value.get(self.name),
+                (str, bytes, _FileCache)):
             if record.id < 0:
-                return ''
-            context = record.context_get()
+                return b''
+            context = record.get_context()
             try:
                 values, = RPCExecute('model', record.model_name, 'read',
                     [record.id], [self.name], context=context)
             except RPCException:
-                return ''
+                return b''
             _, filename = tempfile.mkstemp(prefix='tryton_')
+            data = values[self.name] or b''
+            if isinstance(data, str):
+                data = data.encode('utf-8')
             with open(filename, 'wb') as fp:
-                fp.write(values[self.name] or '')
-            self.set(record, filename)
+                fp.write(data)
+            self.set(record, _FileCache(filename))
         return self.get(record)
 
 
@@ -909,24 +969,98 @@ class DictField(Field):
 
     _default = {}
 
+    def __init__(self, attrs):
+        super(DictField, self).__init__(attrs)
+        self.keys = {}
+
     def get(self, record):
         return super(DictField, self).get(record) or self._default
 
     def get_client(self, record):
         return super(DictField, self).get_client(record) or self._default
 
-    def validation_domains(self, record):
-        screen_domain, attr_domain = self.domains_get(record)
+    def validation_domains(self, record, pre_validate=None):
+        screen_domain, attr_domain = self.domains_get(record, pre_validate)
         return screen_domain
 
     def domain_get(self, record):
         screen_domain, attr_domain = self.domains_get(record)
-        return concat(localize_domain(inverse_leaf(screen_domain)),
-            attr_domain)
+        return concat(localize_domain(screen_domain), attr_domain)
+
+    def date_format(self, record):
+        context = self.get_context(record)
+        return common.date_format(context.get('date_format'))
+
+    def time_format(self, record):
+        return '%X'
+
+    def add_keys(self, keys, record):
+        schema_model = self.attrs['schema_model']
+        context = self.get_context(record)
+        domain = self.domain_get(record)
+        batchlen = min(10, CONFIG['client.limit'])
+        for i in range(0, len(keys), batchlen):
+            sub_keys = keys[i:i + batchlen]
+            try:
+                key_ids = RPCExecute('model', schema_model, 'search',
+                    [('name', 'in', sub_keys), domain], 0,
+                    CONFIG['client.limit'], None, context=context)
+            except RPCException:
+                key_ids = []
+            if not key_ids:
+                continue
+            try:
+                values = RPCExecute('model', schema_model,
+                    'get_keys', key_ids, context=context)
+            except RPCException:
+                values = []
+            if not values:
+                continue
+            self.keys.update({k['name']: k for k in values})
+
+    def add_new_keys(self, key_ids, record):
+        schema_model = self.attrs['schema_model']
+        context = self.get_context(record)
+        try:
+            new_fields = RPCExecute('model', schema_model,
+                'get_keys', key_ids, context=context)
+        except RPCException:
+            new_fields = []
+
+        new_keys = []
+        for new_field in new_fields:
+            name = new_field['name']
+            new_keys.append(name)
+            self.keys[name] = new_field
+
+        return new_keys
+
+    def validate(self, record, softvalidation=False, pre_validate=None):
+        valid = super(DictField, self).validate(
+            record, softvalidation, pre_validate)
+
+        if self.attrs.get('readonly'):
+            return valid
+
+        decoder = PYSONDecoder()
+        field_value = self.get_eval(record)
+        domain = []
+        for key in field_value:
+            if key not in self.keys:
+                continue
+            key_domain = self.keys[key].get('domain')
+            if key_domain:
+                domain.append(decoder.decode(key_domain))
+
+        valid_value = eval_domain(domain, field_value)
+        if not valid_value:
+            self.get_state_attrs(record)['invalid'] = 'domain'
+
+        return valid and valid_value
+
 
 TYPES = {
     'char': CharField,
-    'float_time': FloatField,
     'integer': IntegerField,
     'biginteger': IntegerField,
     'float': FloatField,
@@ -940,6 +1074,8 @@ TYPES = {
     'datetime': DateTimeField,
     'date': DateField,
     'time': TimeField,
+    'timestamp': DateTimeField,
+    'timedelta': TimeDeltaField,
     'one2one': O2OField,
     'binary': BinaryField,
     'dict': DictField,
