@@ -256,7 +256,7 @@ class Screen(SignalEvent):
                 self.clear()
                 self.context_screen.display(set_cursor=True)
                 return False
-            context = self.context
+            context = self.local_context
             context.update(self.context_screen.get_on_change_value())
             self.new_group(context)
 
@@ -307,6 +307,7 @@ class Screen(SignalEvent):
                 domain = self.domain_parser.parse(search_string)
             else:
                 domain = self.search_value
+                self.search_value = None
             if set_text:
                 self.screen_container.set_text(
                     self.domain_parser.string(domain))
@@ -356,6 +357,13 @@ class Screen(SignalEvent):
             context['context_model'] = self.context_screen.model_name
         return context
 
+    @property
+    def local_context(self):
+        context = self.group.local_context
+        if self.context_screen:
+            context['context_model'] = self.context_screen.model_name
+        return context
+
     def __get_group(self):
         return self.__group
 
@@ -397,8 +405,8 @@ class Screen(SignalEvent):
 
     def _group_cleared(self, group, signal):
         for view in self.views:
-            if hasattr(view, 'reload'):
-                view.reload = True
+            if view.view_type == 'tree':
+                view.display(force=True)
 
     def _group_list_changed(self, group, signal):
         for view in self.views:
@@ -478,7 +486,7 @@ class Screen(SignalEvent):
                 'model': self.model_name,
                 'id': self.current_record.id if self.current_record else None,
                 'ids': [r.id for r in self.selected_records],
-                }, context=self.group._context.copy(), warning=False)
+                }, context=self.local_context, warning=False)
         else:
             if not self.modified():
                 self.switch_view(view_type='form')
@@ -1204,7 +1212,7 @@ class Screen(SignalEvent):
         if self.domain:
             query_string.append(('domain', json.dumps(
                         self.domain, cls=JSONEncoder, separators=(',', ':'))))
-        context = self.group._context  # Avoid rpc context
+        context = self.local_context  # Avoid rpc context
         if context:
             query_string.append(('context', json.dumps(
                         context, cls=JSONEncoder, separators=(',', ':'))))
@@ -1217,8 +1225,11 @@ class Screen(SignalEvent):
         path = [CONFIG['login.db'], 'model', self.model_name]
         view_ids = [v.view_id for v in self.views] + self.view_ids
         if self.current_view.view_type != 'form':
-            search_string = self.screen_container.get_text()
-            search_value = self.domain_parser.parse(search_string)
+            if self.search_value:
+                search_value = self.search_value
+            else:
+                search_string = self.screen_container.get_text()
+                search_value = self.domain_parser.parse(search_string)
             if search_value:
                 query_string.append(('search_value', json.dumps(
                             search_value, cls=JSONEncoder,
