@@ -1,52 +1,55 @@
-#This file is part of Tryton.  The COPYRIGHT file at the top level of
-#this repository contains the full copyright notices and license terms.
+# This file is part of Tryton.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
 import gettext
 
 import gobject
 import gtk
-from interface import WidgetInterface, TranslateMixin
-from tryton.common import Tooltips
-from tryton.common.entry_position import manage_entry_position
+from .widget import Widget, TranslateMixin
+from tryton.common import Tooltips, IconFactory
+from tryton.common.entry_position import reset_position
+from tryton.common.selection import PopdownMixin, selection_shortcuts
+from tryton.config import CONFIG
 
 _ = gettext.gettext
 
 
-class Char(WidgetInterface, TranslateMixin):
+class Char(Widget, TranslateMixin, PopdownMixin):
     "Char"
 
-    def __init__(self, field_name, model_name, attrs=None):
-        super(Char, self).__init__(field_name, model_name, attrs=attrs)
+    def __init__(self, view, attrs):
+        super(Char, self).__init__(view, attrs)
 
         self.widget = gtk.HBox()
         self.autocomplete = bool(attrs.get('autocomplete'))
         if self.autocomplete:
             self.entry = gtk.ComboBoxEntry()
-            self.entry_store = gtk.ListStore(gobject.TYPE_STRING)
-            self.entry.set_model(self.entry_store)
-            self.entry.set_text_column(0)
-            completion = gtk.EntryCompletion()
-            completion.set_model(self.entry_store)
-            completion.set_text_column(0)
-            self.entry.get_child().set_completion(completion)
+            selection_shortcuts(self.entry)
             focus_entry = self.entry.get_child()
+            self.set_popdown([], self.entry)
+            self.entry.connect('changed', self.changed)
+            self.entry.connect('move-active', self._move_active)
+            self.entry.connect(
+                'scroll-event',
+                lambda c, e: c.emit_stop_by_name('scroll-event'))
         else:
             self.entry = gtk.Entry()
             focus_entry = self.entry
+        self.mnemonic_widget = focus_entry
 
         focus_entry.set_property('activates_default', True)
         focus_entry.connect('activate', self.sig_activate)
         focus_entry.connect('focus-out-event', lambda x, y: self._focus_out())
         focus_entry.connect('key-press-event', self.send_modified)
-        manage_entry_position(focus_entry)
         expand, fill = True, True
         if attrs.get('size'):
             expand, fill = False, False
         self.widget.pack_start(self.entry, expand=expand, fill=fill)
 
-        self.button = None
         if attrs.get('translate'):
-            self.button = self.translate_button()
-            self.widget.pack_start(self.button, False, False)
+            self.entry.set_icon_from_pixbuf(
+                gtk.ENTRY_ICON_SECONDARY,
+                IconFactory.get_pixbuf('tryton-translate', gtk.ICON_SIZE_MENU))
+            self.entry.connect('icon-press', self.translate)
 
     def translate_widget(self):
         entry = gtk.Entry()
@@ -57,33 +60,31 @@ class Char(WidgetInterface, TranslateMixin):
             entry.set_max_length(field_size or 0)
         return entry
 
-    @staticmethod
-    def translate_widget_set(widget, value):
-        widget.set_text(value or '')
+    def translate_widget_set(self, widget, value):
+        widget.set_text(value)
+        reset_position(widget)
 
-    @staticmethod
-    def translate_widget_get(widget):
+    def translate_widget_get(self, widget):
         return widget.get_text()
 
-    @staticmethod
-    def translate_widget_set_readonly(widget, value):
+    def translate_widget_set_readonly(self, widget, value):
         widget.set_editable(not value)
         widget.props.sensitive = not value
 
-    def _color_widget(self):
-        if self.autocomplete:
-            return self.entry.get_child()
-        return self.entry
-
-    def grab_focus(self):
-        return self.entry.grab_focus()
+    def changed(self, combobox):
+        def focus_out():
+            if combobox.props.window:
+                self._focus_out()
+        # Only when changed from pop list
+        if not combobox.get_child().has_focus():
+            # Must be deferred because it triggers a display of the form
+            gobject.idle_add(focus_out)
 
     @property
     def modified(self):
         if self.record and self.field:
-            entry = self.entry.get_child() if self.autocomplete else self.entry
-            value = entry.get_text() or ''
-            return self.field.get_client(self.record) != value
+            value = self.get_client_value(self.record, self.field)
+            return value != self.get_value()
         return False
 
     def set_value(self, record, field):
@@ -95,17 +96,23 @@ class Char(WidgetInterface, TranslateMixin):
         entry = self.entry.get_child() if self.autocomplete else self.entry
         return entry.get_text()
 
+    def get_client_value(self, record, field):
+        if not field:
+            value = ''
+        else:
+            value = field.get_client(record)
+        return value
+
     def display(self, record, field):
         super(Char, self).display(record, field)
-        if record and self.autocomplete:
-            autocompletion = record.autocompletion.get(self.field_name, [])
-            current = [elem[0] for elem in self.entry_store]
-            if autocompletion != current:
-                self.entry_store.clear()
-                for row in autocompletion:
-                    self.entry_store.append((row,))
-        elif self.autocomplete:
-            self.entry_store.clear()
+        if self.autocomplete:
+            if record:
+                if self.field_name not in record.autocompletion:
+                    record.do_autocomplete(self.field_name)
+                selection = record.autocompletion.get(self.field_name, [])
+            else:
+                selection = []
+            self.set_popdown([(x, x) for x in selection], self.entry)
 
         # Set size
         if self.autocomplete:
@@ -120,41 +127,41 @@ class Char(WidgetInterface, TranslateMixin):
             size_entry.set_width_chars(-1)
             size_entry.set_max_length(0)
 
-        if not field:
-            value = ''
-        else:
-            value = field.get_client(record)
-
+        value = self.get_client_value(record, field)
         if not self.autocomplete:
             self.entry.set_text(value)
+            reset_position(self.entry)
         else:
-            for idx, row in enumerate(self.entry_store):
-                if row[0] == value:
-                    self.entry.set_active(idx)
-                    return
-            else:
-                self.entry.get_child().set_text(value)
+            self.entry.handler_block_by_func(self.changed)
+            if not self.set_popdown_value(self.entry, value) or not value:
+                child = self.entry.get_child()
+                child.set_text(value)
+                reset_position(child)
+            self.entry.handler_unblock_by_func(self.changed)
+
+    def _move_active(self, combobox, scroll_type):
+        if not combobox.get_child().get_editable():
+            combobox.emit_stop_by_name('move-active')
 
     def _readonly_set(self, value):
         sensitivity = {True: gtk.SENSITIVITY_OFF, False: gtk.SENSITIVITY_AUTO}
         super(Char, self)._readonly_set(value)
         if self.autocomplete:
-            self.entry.get_child().set_editable(not value)
+            entry_editable = self.entry.get_child()
             self.entry.set_button_sensitivity(sensitivity[value])
         else:
-            self.entry.set_editable(not value)
-        if self.button:
-            self.button.set_sensitive(not value)
-        if value:
+            entry_editable = self.entry
+        entry_editable.set_editable(not value)
+        if value and CONFIG['client.fast_tabbing']:
             self.widget.set_focus_chain([])
         else:
-            self.widget.set_focus_chain([self.entry])
+            self.widget.unset_focus_chain()
 
 
 class Password(Char):
 
-    def __init__(self, field_name, model_name, attrs=None):
-        super(Password, self).__init__(field_name, model_name, attrs=attrs)
+    def __init__(self, view, attrs):
+        super(Password, self).__init__(view, attrs)
         self.entry.props.visibility = False
 
         self.visibility_checkbox = gtk.CheckButton()
@@ -166,10 +173,10 @@ class Password(Char):
         super(Char, self)._readonly_set(value)
         self.entry.set_editable(not value)
         self.visibility_checkbox.props.visible = not value
-        if value:
+        if value and CONFIG['client.fast_tabbing']:
             self.widget.set_focus_chain([])
         else:
-            self.widget.set_focus_chain([self.entry, self.visibility_checkbox])
+            self.widget.unset_focus_chain()
 
     def toggle_visibility(self, button):
         self.entry.props.visibility = not self.entry.props.visibility
