@@ -9,15 +9,19 @@ import decimal
 import gettext
 from decimal import Decimal
 
-from interface import WidgetInterface
-from tryton.config import CONFIG
+from .widget import Widget
 from tryton.gui.window.win_search import WinSearch
-from tryton.common import RPCExecute, RPCException, timezoned_date, \
-    datetime_strftime, Tooltips
-from tryton.common.date_widget import DateEntry
-from tryton.common.placeholder_entry import PlaceholderEntry
+from tryton.common import Tooltips, timezoned_date, untimezoned_date, \
+        IconFactory
 from tryton.common.selection import selection_shortcuts
-from tryton.translate import date_format
+from tryton.common.completion import get_completion, update_completion
+from tryton.common.datetime_ import Date, DateTime
+from tryton.common.domain_parser import quote
+from tryton.common.entry_position import reset_position
+from tryton.common.underline import set_underline
+from tryton.common.domain_inversion import eval_domain
+from tryton.common.widget_style import widget_class
+from tryton.pyson import PYSONDecoder
 
 _ = gettext.gettext
 
@@ -28,7 +32,7 @@ class DictEntry(object):
 
     def __init__(self, name, parent_widget):
         self.name = name
-        self.definition = parent_widget.keys[name]
+        self.definition = parent_widget.field.keys[name]
         self.parent_widget = parent_widget
         self.widget = self.create_widget()
 
@@ -37,6 +41,8 @@ class DictEntry(object):
         widget.connect('key-press-event', self.parent_widget.send_modified)
         widget.connect('focus-out-event',
             lambda w, e: self.parent_widget._focus_out())
+        widget.props.activates_default = True
+        widget.connect('activate', self.parent_widget.sig_activate)
         return widget
 
     def modified(self, value):
@@ -46,7 +52,8 @@ class DictEntry(object):
         return self.widget.get_text()
 
     def set_value(self, value):
-        return self.widget.set_text(value or '')
+        self.widget.set_text(value or '')
+        reset_position(self.widget)
 
     def set_readonly(self, readonly):
         self.widget.set_editable(not readonly)
@@ -56,7 +63,7 @@ class DictBooleanEntry(DictEntry):
 
     def create_widget(self):
         widget = gtk.CheckButton()
-        widget.connect('toggled', self.parent_widget.send_modified)
+        widget.connect('toggled', self.parent_widget.sig_activate)
         widget.connect('focus-out-event', lambda w, e:
             self.parent_widget._focus_out())
         return widget
@@ -65,7 +72,12 @@ class DictBooleanEntry(DictEntry):
         return self.widget.props.active
 
     def set_value(self, value):
-        self.widget.props.active = bool(value)
+        self.widget.handler_block_by_func(self.parent_widget.sig_activate)
+        try:
+            self.widget.props.active = bool(value)
+        finally:
+            self.widget.handler_unblock_by_func(
+                self.parent_widget.sig_activate)
 
     def set_readonly(self, readonly):
         self.widget.set_sensitive(not readonly)
@@ -79,14 +91,17 @@ class DictSelectionEntry(DictEntry):
         widget = gtk.ComboBoxEntry()
 
         # customizing entry
-        widget.child.props.activates_default = True
-        widget.child.connect('changed', self.parent_widget.send_modified)
-        widget.child.connect('focus-out-event',
+        child = widget.get_child()
+        child.props.activates_default = True
+        child.connect('changed', self.parent_widget.send_modified)
+        child.connect('focus-out-event',
             lambda w, e: self.parent_widget._focus_out())
-        widget.child.connect('activate',
+        child.connect('activate',
             lambda w: self.parent_widget._focus_out())
         widget.connect('notify::active',
             lambda w, e: self.parent_widget._focus_out())
+        widget.connect(
+            'scroll-event', lambda c, e: c.emit_stop_by_name('scroll-event'))
         selection_shortcuts(widget)
 
         # setting completion and selection
@@ -94,29 +109,32 @@ class DictSelectionEntry(DictEntry):
         model.append(('',))
         self._selection = {'': None}
         width = 10
-        for value, name in sorted(self.definition['selection'],
-                key=operator.itemgetter(1)):
+        selection = self.definition['selection']
+        if self.definition.get('sorted', True):
+            selection.sort(key=operator.itemgetter(1))
+        for value, name in selection:
             name = str(name)
             self._selection[name] = value
             model.append((name,))
             width = max(width, len(name))
         widget.set_model(model)
         widget.set_text_column(0)
-        widget.child.set_width_chars(width)
+        child.set_width_chars(width)
         completion = gtk.EntryCompletion()
         completion.set_inline_selection(True)
         completion.set_model(model)
-        widget.child.set_completion(completion)
+        child.set_completion(completion)
         completion.set_text_column(0)
         return widget
 
     def get_value(self):
-        if not self.widget.child:  # widget is destroyed
+        child = self.widget.get_child()
+        if not child:  # widget is destroyed
             return
-        text = self.widget.child.get_text()
+        text = child.get_text()
         value = None
         if text:
-            for txt, val in self._selection.items():
+            for txt, val in list(self._selection.items()):
                 if not val:
                     continue
                 if txt[:len(text)].lower() == text.lower():
@@ -127,7 +145,9 @@ class DictSelectionEntry(DictEntry):
 
     def set_value(self, value):
         values = dict(self.definition['selection'])
-        self.widget.child.set_text(values.get(value, ''))
+        child = self.widget.get_child()
+        child.set_text(values.get(value, ''))
+        reset_position(child)
 
     def set_readonly(self, readonly):
         self.widget.set_sensitive(not readonly)
@@ -171,6 +191,7 @@ class DictIntegerEntry(DictEntry):
         else:
             txt_val = ''
         self.widget.set_text(txt_val)
+        reset_position(self.widget)
 
 
 class DictFloatEntry(DictIntegerEntry):
@@ -227,6 +248,7 @@ class DictFloatEntry(DictIntegerEntry):
             txt_val = ''
         self.widget.set_width_chars(sum(digits))
         self.widget.set_text(txt_val)
+        reset_position(self.widget)
 
 
 class DictNumericEntry(DictFloatEntry):
@@ -246,50 +268,45 @@ class DictDateTimeEntry(DictEntry):
     fill = False
 
     def create_widget(self):
-        widget = DateEntry('')
-        widget.set_format(self.get_format())
+        widget = DateTime()
+        record = self.parent_widget.record
+        field = self.parent_widget.field
+        if record and field:
+            format_ = field.time_format(record)
+            widget.props.format = format_
         widget.connect('key_press_event', self.parent_widget.send_modified)
         widget.connect('focus-out-event', lambda w, e:
             self.parent_widget._focus_out())
         return widget
 
-    def modified(self, value):
-        if value.get(self.name):
-            text = datetime_strftime(timezoned_date(value[self.name]),
-                self.get_format())
-        else:
-            text = ''
-        return self.widget.compute_date(self.widget.get_text()) != text
-
-    def get_format(self):
-        return date_format() + ' %H:%M:%S'  # got to find a way
-
     def get_value(self):
-        return self.widget.date_get()
+        return untimezoned_date(self.widget.props.value)
 
     def set_value(self, value):
-        self.widget.date_set(value)
-        txt = self.widget.get_text()
-        if txt:
-            if len(txt) > self.widget.get_width_chars():
-                self.widget.set_width_chars(len(txt))
+        self.widget.props.value = timezoned_date(value)
 
 
-class DictDateEntry(DictDateTimeEntry):
+class DictDateEntry(DictEntry):
+    expand = False
+    fill = False
 
-    def modified(self, value):
-        if value.get(self.name):
-            text = datetime_strftime(value[self.name], self.get_format())
-        else:
-            text = ''
-        return self.widget.compute_date(self.widget.get_text()) != text
-
-    def get_format(self):
-        return date_format()
+    def create_widget(self):
+        widget = Date()
+        record = self.parent_widget.record
+        field = self.parent_widget.field
+        if record and field:
+            format_ = field.date_format(record)
+            widget.props.format = format_
+        widget.connect('key_press_event', self.parent_widget.send_modified)
+        widget.connect('focus-out-event', lambda w, e:
+            self.parent_widget._focus_out())
+        return widget
 
     def get_value(self):
-        dt = super(DictDateEntry, self).get_value()
-        return dt.date() if dt else None
+        return self.widget.props.value
+
+    def set_value(self, value):
+        self.widget.props.value = value
 
 
 DICT_ENTRIES = {
@@ -304,17 +321,19 @@ DICT_ENTRIES = {
     }
 
 
-class DictWidget(WidgetInterface):
+class DictWidget(Widget):
 
-    def __init__(self, field_name, model_name, attrs=None):
-        super(DictWidget, self).__init__(field_name, model_name, attrs=attrs)
+    def __init__(self, view, attrs):
+        super(DictWidget, self).__init__(view, attrs)
         self.schema_model = attrs['schema_model']
-        self.keys = {}
         self.fields = {}
         self.buttons = {}
         self.rows = {}
 
-        self.widget = gtk.Frame(attrs.get('string', ''))
+        self.widget = gtk.Frame()
+        label = gtk.Label(set_underline(attrs.get('string', '')))
+        label.set_use_underline(True)
+        self.widget.set_label_widget(label)
         self.widget.set_shadow_type(gtk.SHADOW_OUT)
 
         vbox = gtk.VBox()
@@ -328,17 +347,26 @@ class DictWidget(WidgetInterface):
 
         hbox = gtk.HBox()
         hbox.set_border_width(2)
-        self.wid_text = PlaceholderEntry()
+        self.wid_text = gtk.Entry()
         self.wid_text.set_placeholder_text(_('Search'))
         self.wid_text.props.width_chars = 13
         self.wid_text.connect('activate', self._sig_activate)
         hbox.pack_start(self.wid_text, expand=True, fill=True)
+        label.set_mnemonic_widget(self.wid_text)
+
+        if int(self.attrs.get('completion', 1)):
+            self.wid_completion = get_completion(search=False, create=False)
+            self.wid_completion.connect('match-selected',
+                self._completion_match_selected)
+            self.wid_text.set_completion(self.wid_completion)
+            self.wid_text.connect('changed', self._update_completion)
+        else:
+            self.wid_completion = None
+
         self.but_add = gtk.Button()
         self.but_add.connect('clicked', self._sig_add)
-        img_add = gtk.Image()
-        img_add.set_from_stock('tryton-list-add', gtk.ICON_SIZE_SMALL_TOOLBAR)
-        img_add.set_alignment(0.5, 0.5)
-        self.but_add.add(img_add)
+        self.but_add.add(IconFactory.get_image(
+                'tryton-add', gtk.ICON_SIZE_SMALL_TOOLBAR))
         self.but_add.set_relief(gtk.RELIEF_NONE)
         hbox.pack_start(self.but_add, expand=False, fill=False)
         hbox.set_focus_chain([self.wid_text])
@@ -351,13 +379,14 @@ class DictWidget(WidgetInterface):
         self._readonly = False
         self._record_id = None
 
+    @property
+    def _invalid_widget(self):
+        return self.wid_text
+
     def _new_remove_btn(self):
         but_remove = gtk.Button()
-        img_remove = gtk.Image()
-        img_remove.set_from_stock('tryton-list-remove',
-            gtk.ICON_SIZE_SMALL_TOOLBAR)
-        img_remove.set_alignment(0.5, 0.5)
-        but_remove.add(img_remove)
+        but_remove.add(IconFactory.get_image(
+                'tryton-remove', gtk.ICON_SIZE_SMALL_TOOLBAR))
         but_remove.set_relief(gtk.RELIEF_NONE)
         return but_remove
 
@@ -366,38 +395,32 @@ class DictWidget(WidgetInterface):
             self._sig_add()
 
     def _sig_add(self, *args):
-        context = self.field.context_get(self.record)
+        context = self.field.get_context(self.record)
         value = self.wid_text.get_text()
         domain = self.field.domain_get(self.record)
-        dom = [('rec_name', 'ilike', '%%%s%%' % value)] if value else []
-        dom.append(('id', 'not in',
-                [self.keys[f]['id'] for f in self.fields]))
-        try:
-            ids = RPCExecute('model', self.schema_model, 'search',
-                domain + dom, 0, CONFIG['client.limit'], None, context=context)
-        except RPCException:
-            return False
 
         def callback(result):
             if result:
-                self.send_modified()
-                try:
-                    new_fields = RPCExecute('model', self.schema_model,
-                        'get_keys', [r[0] for r in result],
-                        context=context)
-                except RPCException:
-                    new_fields = []
-                for new_field in new_fields:
-                    if new_field['name'] not in self.fields:
-                        self.keys[new_field['name']] = new_field
-                        self.add_line(new_field['name'])
+                self.add_new_keys([r[0] for r in result])
             self.wid_text.set_text('')
 
-        if len(ids) != 1:
-            WinSearch(self.schema_model, callback, sel_multi=True, ids=ids,
-                context=context, domain=domain, new=False)
-        else:
-            callback([(id, None) for id in ids])
+        win = WinSearch(self.schema_model, callback, sel_multi=True,
+            context=context, domain=domain, new=False)
+        win.screen.search_filter(quote(value))
+        win.show()
+
+    def add_new_keys(self, ids):
+        new_keys = self.field.add_new_keys(ids, self.record)
+        self.send_modified()
+        focus = False
+        for key_name in new_keys:
+            if key_name not in self.fields:
+                self.add_line(key_name)
+                if not focus:
+                    # Use idle add because it can be called from the callback
+                    # of WinSearch while the popup is still there
+                    gobject.idle_add(self.fields[key_name].widget.grab_focus)
+                    focus = True
 
     def _sig_remove(self, button, key, modified=True):
         del self.fields[key]
@@ -415,34 +438,36 @@ class DictWidget(WidgetInterface):
 
     def get_value(self):
         return dict((key, widget.get_value())
-            for key, widget in self.fields.items())
+            for key, widget in list(self.fields.items()))
 
     @property
     def modified(self):
         if self.record and self.field:
             value = self.field.get_client(self.record)
             return any(widget.modified(value)
-                for widget in self.fields.itervalues())
+                for widget in self.fields.values())
         return False
 
     def _readonly_set(self, readonly):
         self._readonly = readonly
         self._set_button_sensitive()
-        for widget in self.fields.values():
+        for widget in list(self.fields.values()):
             widget.set_readonly(readonly)
+        self.wid_text.set_sensitive(not readonly)
         self.wid_text.set_editable(not readonly)
 
     def _set_button_sensitive(self):
         self.but_add.set_sensitive(bool(
                 not self._readonly
                 and self.attrs.get('create', True)))
-        for button in self.buttons.itervalues():
+        for button in self.buttons.values():
             button.set_sensitive(bool(
                     not self._readonly
                     and self.attrs.get('delete', True)))
 
     def add_line(self, key):
-        self.fields[key] = DICT_ENTRIES[self.keys[key]['type_']](key, self)
+        key_schema = self.field.keys[key]
+        self.fields[key] = DICT_ENTRIES[key_schema['type_']](key, self)
         field = self.fields[key]
         alignment = gtk.Alignment(
             float(self.attrs.get('xalign', 0.0)),
@@ -454,21 +479,20 @@ class DictWidget(WidgetInterface):
         alignment.add(hbox)
         n_rows = self.table.props.n_rows
         self.table.resize(n_rows + 1, 3)
-        if gtk.widget_get_default_direction() == gtk.TEXT_DIR_RTL:
-            text = _(':') + self.keys[key]['string']
-        else:
-            text = self.keys[key]['string'] + _(':')
-        label = gtk.Label(text)
+        text = key_schema['string'] + _(':')
+        label = gtk.Label(set_underline(text))
+        label.set_use_underline(True)
         label.set_alignment(1., .5)
         self.table.attach(label, 0, 1, n_rows - 1, n_rows,
             xoptions=gtk.FILL, yoptions=False, xpadding=2)
+        label.set_mnemonic_widget(field.widget)
         label.show()
         self.table.attach(alignment, 1, 2, n_rows - 1, n_rows,
             xoptions=gtk.FILL | gtk.EXPAND, yoptions=False, xpadding=2)
         alignment.show_all()
         remove_but = self._new_remove_btn()
         self.tooltips.set_tip(remove_but, _('Remove "%s"') %
-            self.keys[key]['string'])
+            key_schema['string'])
         self.table.attach(remove_but, 2, 3, n_rows - 1, n_rows,
             xoptions=gtk.FILL, yoptions=False, xpadding=2)
         remove_but.connect('clicked', self._sig_remove, key)
@@ -476,17 +500,6 @@ class DictWidget(WidgetInterface):
         self.rows[key] = [label, alignment, remove_but]
         self.buttons[key] = remove_but
 
-    def add_key(self, key):
-        context = self.field.context_get(self.record)
-        try:
-            key_ids = RPCExecute('model', self.schema_model, 'search',
-                [('name', '=', key)], 0, CONFIG['client.limit'],
-                None, context=context)
-            self.keys[key] = RPCExecute('model', self.schema_model,
-                'get_keys', key_ids, context=context)[0]
-        except RPCException:
-            pass
-
     def display(self, record, field):
         super(DictWidget, self).display(record, field)
 
@@ -495,21 +508,46 @@ class DictWidget(WidgetInterface):
 
         record_id = record.id if record else None
         if record_id != self._record_id:
-            for key in self.fields.keys():
+            for key in list(self.fields.keys()):
                 self._sig_remove(None, key, modified=False)
             self._record_id = record_id
 
         value = field.get_client(record) if field else {}
-        for key in sorted(value.iterkeys()):
-            val = value[key]
-            if key not in self.keys:
-                self.add_key(key)
+        new_key_names = set(value.keys()) - set(field.keys)
+        if new_key_names:
+            field.add_keys(list(new_key_names), self.record)
+        decoder = PYSONDecoder()
+        for key, val in sorted(value.items()):
+            if key not in field.keys:
+                continue
             if key not in self.fields:
                 self.add_line(key)
             widget = self.fields[key]
             widget.set_value(val)
             widget.set_readonly(self._readonly)
+            key_domain = decoder.decode(
+                self.field.keys[key].get('domain') or '[]')
+            widget_class(
+                widget.widget, 'invalid', not eval_domain(key_domain, value))
         for key in set(self.fields.keys()) - set(value.keys()):
             self._sig_remove(None, key, modified=False)
 
         self._set_button_sensitive()
+
+    def _completion_match_selected(self, completion, model, iter_):
+        record_id, = model.get(iter_, 1)
+        self.add_new_keys([record_id])
+        self.wid_text.set_text('')
+
+        completion_model = self.wid_completion.get_model()
+        completion_model.clear()
+        completion_model.search_text = self.wid_text.get_text()
+        return True
+
+    def _update_completion(self, widget):
+        if not self.wid_text.get_editable():
+            return
+        if not self.record:
+            return
+        update_completion(self.wid_text, self.record, self.field,
+            self.schema_model)
